#include <emc/io.h>
#include <emc/rate.h>
#include <emc/odom.h>
#include <cmath>

#include "driveControl.h"
#include "detection.h"
#include "worldModel.h"

#include "config.h"
#include <iostream>

typedef enum {
    stop = 0,
    moveAwayWall,
    fullRotScan,
    locateExit,
    prepareWallFollow,
    rotateExitEntrance,
    positionExitEntrance,
    driveToExitEntrance,
    alignFinish,
    finish,
    enterExit,
    test,
    turnAround,
    turnado,
    //
    wallFollow
} state_t;
using namespace std;

int main(int argc, char *argv[])
{
    // Initialization of Robot
    emc::Rate r(EXECUTION_RATE);
    emc::IO io;
    emc::OdometryData odom;

    // Initialize the Classes
    DriveControl pico_drive(&io);
    Detection detection(&io);
    WorldModel worldModel;

    // Initialize the State of the State Machine
    //state_t state = rotate_exit;//exit_detect;
    state_t state_backup = wallFollow;
    state_t state = moveAwayWall;
    //state = test;
    //state = rotateExitEntrance;
    //state = turnAround;

    /* Initialising some "memory" */
    double toRotate = 0;
    double rotatedAngle = 0.0;
    int iExitCheck = 0;
    bool lookedAround = 0;
    double desiredRotation = 0;
    double drivenDistance = 0;
    double distanceBackwards = 0;
    /* Main Execution Loop, this loop keeps running until io.ok returns false,
     * hence on a robot error. */
    while(io.ok()) {
        // Get the Sensor data from the LRF
        if(detection.getSensorData()) { //necessary?
            // Feed the WorldModel
            switch(state) {
            /*
            case test:
                    pico_drive.rotateSet(ROTATE_SPEED,1);
                    state = positionExitEntrance;
                    break;*/
            case stop:
                pico_drive.stop();
                break;

            case moveAwayWall: //for rare scenario (back against the wall)

/*
                //if(*worldModel.getMinimumDistance() < OPTIMAL_DISTANCE){//too close to wall for comfort!
                    distanceBackwards += pico_drive.driveBackward(FORWARD_SPEED);
                    if(fabs(distanceBackwards) >= 2*OPTIMAL_DISTANCE) {
                    distanceBackwards=0;
                    //state = fullRotScan;
                    }
                //}*/
                state = locateExit;
                break;

            case locateExit: //0: exit free in front, 1: exit located through corner, 2: no exit found
                iExitCheck = worldModel.findExit(&(detection.laser));
                if(iExitCheck==0){
                    state = rotateExitEntrance;
                    //pico_drive.driveForward(2);
                }else if(iExitCheck==1){
                     state = rotateExitEntrance;
                }
                else if(iExitCheck==2&&lookedAround==0){ //LOOK BEHIND
                    lookedAround = 1; //dont keep backchecking
                    desiredRotation = 1*M_PI;
                    state_backup = locateExit;
                    state = turnado;
                }else state = prepareWallFollow;
                break;

            case turnado:
                if(desiredRotation >= 0.5*M_PI){
                    cout << " \n turning big angle";
                    rotatedAngle += pico_drive.rotate(ROTATE_SPEED);
                    if(fabs(rotatedAngle) >= 0.5*M_PI) {
                        rotatedAngle = 0.0;
                        desiredRotation = desiredRotation - 0.5*M_PI;
                        std::cout << "\n\nTOGO: " << desiredRotation;
                        state = turnado;
                    }
                }else if(desiredRotation <= 0.5*M_PI){
                    cout << "\n small angle turning";
                    rotatedAngle += pico_drive.rotate(ROTATE_SPEED);
                    if(fabs(rotatedAngle) >= desiredRotation) {
                        // done turning
                        rotatedAngle = 0.0;
                        desiredRotation = 0;
                        pico_drive.rotate(0);
                        state = state_backup;
                    }
                }
                    break;

            case rotateExitEntrance: //rotate for direct line to exit entrnace
                desiredRotation = *worldModel.getExitAngle();
                if(desiredRotation < 0.1)state = driveToExitEntrance;
                else{
                    state_backup = state;
                    state = turnado;
                }

                break;
            case driveToExitEntrance: //drive ti exit etrance
                drivenDistance += pico_drive.driveDistanceForward(FORWARD_SPEED);
                cout << "\n" << drivenDistance;
                //cout << " driving on! ";
                if(fabs(drivenDistance) >= *worldModel.getExitDistance()){
                    pico_drive.driveForward(0);
                    drivenDistance = 0;
                    state = positionExitEntrance;
                }
                break;

            case enterExit:
                pico_drive.driveForward(1);
                break;

            case positionExitEntrance:
                break;

            case  prepareWallFollow:
                pico_drive.driveForward(10);
                break;

            default:
                pico_drive.stop();
                break;
            }
        } else {
             pico_drive.stop();
        }
        // Use this to ensure an execution rate of 20 Hertz.
        r.sleep();
    }


    return 0;
}
